//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "main.h"
#include "xUtils.hpp"
#include "about.h"
#include "xFiles.hpp"
#include "util.h"
#include "mapunit.h"
#include "sprite.h"
#include "tileunit.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TMainForm *MainForm;
//---------------------------------------------------------------------------
__fastcall TMainForm::TMainForm(TComponent* Owner)
        : TForm(Owner)
{
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::mnuAboutClick(TObject *Sender)
{
  TAboutBox* box = new TAboutBox(this);
  try {
    box->ShowModal();
  } __finally {
    delete box;
  }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::mnuExitClick(TObject *Sender)
{
  Close();        
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::mnuRestoreLevelClick(TObject *Sender)
{
  LevelNo = LevelNo;        
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::mnuChooseLevelClick(TObject *Sender)
{
  if (!AskSaveMap()) return;

  AnsiString S = IntToStr(FLevelNo);
  if (InputQuery(Application->Title, "Input Level No:", S))
    LevelNo = StrToInt(S);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::mnuSaveClick(TObject *Sender)
{
  MySaveMap();       
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::mnuEditKindClick(TObject *Sender)
{
  TMenuItem* item = dynamic_cast<TMenuItem*>(Sender);
  
  item->Checked = true;
  FEditLayer = item->Tag; // ]wsҦ

  UpdateControlStatus();
  UpdateView();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::mnuRemoveItemClick(TObject *Sender)
{
  if (FSelectionX == -1 || FSelectionY == -1) return;

  if (FEditLayer <= LAYER_MAX)
    TMap::Instance().GetCell(FEditLayer, FSelectionX, FSelectionY).TileNo = 0;

  UpdateView();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::mnuResetMapClick(TObject *Sender)
{
  TMap& Map = TMap::Instance();

  switch (dynamic_cast<TComponent*>(Sender)->Tag) {
    case 0: if (FEditLayer <= LAYER_MAX) Map.ResetLayer(FEditLayer); break;
    case 1: Map.ResetAllLayers(); break;
  }
  UpdateView();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::mnuShowSomethingClick(TObject *Sender)
{
  TMenuItem* item = dynamic_cast<TMenuItem*>(Sender);

  item->Checked = !item->Checked;
  UpdateView();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::FormCreate(TObject *Sender)
{
  util_init();

  FEditLayer = 0; // w]aνsҦ
  FButtonPressed = mbMiddle; // ܥثeƹS

  FTank = new TMyTank;
  try {
    FTank->LoadBits(); // ŪJDϧ

    TTiles::Instance().LoadFromFile(AppDir + FN_TILE_ARCHIVE); // ŪJϮw
  } catch(Exception& E) {
    ShowMessage(E.Message);
    PostMessage(Handle, WM_CLOSE, 0, 0); // H asynchronous 覡
  }

  // ŪJϮwyz
  ReadComponentResFile(ChangeFileExt(AppDir + FN_TILE_ARCHIVE, ".TVW"), tvwTiles);
  tvwTiles->Selected = tvwTiles->Items->GetFirstNode();

  // According tile num and size, adjust dimension of pnlView
  pnlView->ClientWidth = WORLD_WIDTH;
  pnlView->ClientHeight = WORLD_HEIGHT;

  ClientHeight = pnlView->Height + stbMain->Height;

  FBackBitmap = new Graphics::TBitmap; // wĥ bitmap
  FBackBitmap->Width = WORLD_WIDTH;
  FBackBitmap->Height = WORLD_HEIGHT;

  LevelNo = 1;
  UpdateControlStatus();

  // ƹЦm
  FSelectionX = -1;
  FSelectionY = -1;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::FormClose(TObject *Sender, TCloseAction &Action)
{
  delete FBackBitmap;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::UpdateView()
{
  DrawBackBitmap();
  pbxView->Canvas->Draw(0, 0, FBackBitmap);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::pbxViewPaint(TObject *Sender)
{
  UpdateView();        
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::DrawBackBitmap()
{
  TMap& Map = TMap::Instance();

  if (FEditLayer == LAYER_TERR || !mnuDrawEditLayerOnly->Checked)
    Map.Draw(FBackBitmap->Canvas, LAYER_TERR); // øsaμh
  else { // YSeaμh, N
    FBackBitmap->Canvas->Brush->Color = clBlack;
    FBackBitmap->Canvas->FillRect(FBackBitmap->Canvas->ClipRect);
  }

  if (FEditLayer == LAYER_TERRITEM || !mnuDrawEditLayerOnly->Checked)
     Map.Draw(FBackBitmap->Canvas, LAYER_TERRITEM); // øsaΪh

  if (FEditLayer == LAYER_ITEM || !mnuDrawEditLayerOnly->Checked)
     Map.Draw(FBackBitmap->Canvas, LAYER_ITEM); // øs~h

  // pGϥΪ̷Qݮu, NNueX
  if (mnuShowGrid->Checked) {
    FBackBitmap->Canvas->Pen->Color = clBlack;
    FBackBitmap->Canvas->Pen->Style = psSolid;
    FBackBitmap->Canvas->Pen->Width = 1;

    // eu
    for (int y = 0; y < TILE_NUM_Y; y++) {
      FBackBitmap->Canvas->MoveTo(0, y * TILE_HEIGHT);
      FBackBitmap->Canvas->LineTo(TILE_NUM_X * TILE_WIDTH, y * TILE_HEIGHT);
    }

    // Aeu
    for (int x = 0; x < TILE_NUM_X; x++) {
      FBackBitmap->Canvas->MoveTo(x * TILE_WIDTH, 0);
      FBackBitmap->Canvas->LineTo(x * TILE_WIDTH, TILE_NUM_Y * TILE_HEIGHT);
    }
  }

  // ϥΪ̷Qݤi樫ϰ
  if (mnuShowSpeicalArea->Checked) {
    FBackBitmap->Canvas->Pen->Color = clRed;
    FBackBitmap->Canvas->Pen->Width = 1;
    FBackBitmap->Canvas->Pen->Style = psDash;
    FBackBitmap->Canvas->Brush->Style = bsClear;

    for (int y = 0; y < TILE_NUM_Y; y++)  // v@ˬd
      for (int x = 0; x < TILE_NUM_X; x++) 
        if (!Map.GetCell(LAYER_TERRITEM, x, y).CanPass) {
          // ~إ[WkWe쥪Uu
          FBackBitmap->Canvas->Rectangle(x * TILE_WIDTH, y * TILE_HEIGHT, (x + 1) * TILE_WIDTH - 1, (y + 1) * TILE_HEIGHT - 1);
          FBackBitmap->Canvas->MoveTo((x + 1) * TILE_WIDTH, y * TILE_HEIGHT);
          FBackBitmap->Canvas->LineTo(x * TILE_WIDTH, (y + 1) * TILE_HEIGHT);
        }
  }

  FTank->Draw(FBackBitmap->Canvas); // eXϮ
  if (FEditLayer == LAYER_HITERRITEM || !mnuDrawEditLayerOnly->Checked)
    Map.Draw(FBackBitmap->Canvas, LAYER_HITERRITEM); // øsaΪh
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::UpdateControlStatus()
{
  const AnsiString LayerDesc[5] = {"a", "aΪ", "~", "aΪ", "D"};

  Caption = Format("%s -  %d (s説A: %s)", OPENARRAY(TVarRec,
    (Application->Title, FLevelNo, LayerDesc[FEditLayer])));

  grdTile->Invalidate();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::SetLevelNo(int Value)
{
  TMap& Map = TMap::Instance();

  try {
    // ŪaɤΨm
    Map.LevelNo = Value;
  } catch(...) {
    // YŪaɥ, Mia
    Map.ResetAllLayers();

    // \W]Xlj
    Map.GetCell(LAYER_TERRITEM, 5, 11).TileNo = 10;
    Map.GetCell(LAYER_TERRITEM, 6, 11).TileNo = 10;
    Map.GetCell(LAYER_TERRITEM, 7, 11).TileNo = 10;
    Map.GetCell(LAYER_TERRITEM, 5, 12).TileNo = 10;
    Map.GetCell(LAYER_TERRITEM, 7, 12).TileNo = 10;

    // \WXl
    Map.GetCell(LAYER_TERRITEM, 6, 12).TileNo = 13;
  }

  FTank->PosX = Map.Role_X; // Nm Map 󤤧ۥX
  FTank->PosY = Map.Role_Y;

  FLevelNo = Value;
  FModified = False; // aϩ|ʡ]M:p^

  UpdateControlStatus();
  UpdateView(); // OѤFsCe
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::pbxViewMouseMove(TObject *Sender,
      TShiftState Shift, int X, int Y)
{
  // wg]Xd~F...
  if (X < 0 || X >= TILE_WIDTH * TILE_NUM_X || Y < 0 ||
    Y >= TILE_HEIGHT * TILE_NUM_Y) return;

  FSelectionX = X / TILE_WIDTH; // y, HϮ欰
  FSelectionY = Y / TILE_HEIGHT;

  pbxView->OnMouseDown(Sender, FButtonPressed, Shift, X, Y); // iHԦm񪫫~ĪG

  UpdateView(); // øCe

  // eXثeҿϰ~
  pbxView->Canvas->Pen->Width = 2;
  pbxView->Canvas->Pen->Color = clBlue;
  pbxView->Canvas->Brush->Style = bsClear;
  pbxView->Canvas->Rectangle(TILEWIDTH[FSelectionX], TILEHEIGHT[FSelectionY],
    TILEWIDTH[FSelectionX + GetSelectionWidth()] , TILEHEIGHT[FSelectionY + GetSelectionHeight()]);

  stbMain->SimpleText = Format("m: (%d, %d) jp: (%d, %d)",
    OPENARRAY(TVarRec, (FSelectionX, FSelectionY, GetSelectionWidth(), GetSelectionHeight())));
}
//---------------------------------------------------------------------------
bool __fastcall TMainForm::AskSaveMap()
{
  if (!FModified) return true;

  // T{O_xsثes誺aɮ
  switch (YesNoCancelBox(Format("Do you want to save this map (level %d) ?",
      OPENARRAY(TVarRec, (FLevelNo))))) {
   case mrYes: MySaveMap();
     break;

   case mrNo: break;

   case mrCancel: return false;
  }

  return true;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::FormCloseQuery(TObject *Sender, bool &CanClose)
{
  if (!AskSaveMap()) CanClose = False;        
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::MySaveMap()
{
  if (!FModified) return;
  if (!ValidateMap()) return;

  TMap& Map = TMap::Instance();

  Map.Role_X = FTank->PosX;
  Map.Role_Y = FTank->PosY;
  Map.SaveToFile();
  FModified = false;
}
//---------------------------------------------------------------------------
// ˬdaϬO_Xk
bool __fastcall TMainForm::ValidateMap()
{
  int FlagCount = 0;

  TMap& Map = TMap::Instance();
  
  // ƬݬݦXXl
  for (int y = 0; y < TILE_NUM_Y; y++)
    for (int x = 0; x < TILE_NUM_X; x++)
      if (TTiles::Instance().Tile[Map.GetCell(LAYER_TERRITEM, x, y).TileNo].Attr.Contains(taFlag))
        FlagCount++;

  if (!FlagCount) {
    ShowMessage("SxX, ");
    return false;
  }

  return true;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::pbxViewMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
  FButtonPressed = mbMiddle; // ]w mbMiddle ܨϥΪ̤w}ƹ
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::pbxViewMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
  // OUƹ, tX OnMouseMove ƥBzk
  // ͩԦ]wĪG
  FButtonPressed = Button;

  if (Button == mbMiddle) return; // ƹ䤣

  TMap& Map = TMap::Instance();

  if (Button == mbLeft) { // O]w
    if (!tvwTiles->Selected) return; // Sws

    int No = (int)tvwTiles->Selected->Data; // ϤsժYYs

    // W, WΥkWTBOĤZJXI, F
    if (FEditLayer != LAYER_TERR && FSelectionY == 0 &&
      (FSelectionX == 0 || FSelectionX == TILE_NUM_X / 2 + 1 ||
      FSelectionX == TILE_NUM_X - 1)) return;

    if (FEditLayer <= LAYER_MAX) {
      for (int MY = 0; MY < GetSelectionHeight(); MY++) // Nsa\W
        for (int MX = 0; MX < GetSelectionWidth(); MX++)
          Map.GetCell(FEditLayer, FSelectionX + MX, FSelectionY + MY).TileNo =
          No + TTiles::Instance().Tile[No].XNum * (MY + grdTile->Selection.Top) +
          (MX + grdTile->Selection.Left);
    } else {
      // ⤣iH\biʪaΤW
      if (!Map.GetCell(LAYER_TERRITEM, FSelectionX, FSelectionY).CanPass) return;

      FTank->PosX = TILEWIDTH[FSelectionX]; // ]wDlm
      FTank->PosY = TILEHEIGHT[FSelectionY];
    }
  } else { // kOM
    if (FEditLayer > LAYER_MAX) return; // ⤣βM

    for (int MY = 0; MY < GetSelectionHeight(); MY++) // Ma
      for (int MX = 0; MX < GetSelectionWidth(); MX++)
        Map.GetCell(FEditLayer, FSelectionX + MX, FSelectionY + MY).TileNo = 0;
  }

  FModified = true; // aϤw
  UpdateView(); // saϵe
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::grdTileDrawCell(TObject *Sender, int ACol,
      int ARow, TRect &Rect, TGridDrawState State)
{
  if (FEditLayer > LAYER_MAX) return;
  if (!tvwTiles->Selected) return;

  int No = (int)tvwTiles->Selected->Data;
  if (ARow * grdTile->ColCount + ACol >= TTiles::Instance().Tile[No].XNum * TTiles::Instance().Tile[No].YNum) return;

  No += ARow * grdTile->ColCount + ACol;
  if (No >= TTiles::Instance().TileNum) return;

  // NϧεeX
  grdTile->Canvas->Draw(Rect.Left, Rect.Top, TTiles::Instance().Tile[No].Bitmap);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::tvwTilesChange(TObject *Sender, TTreeNode *Node)
{
  if (!tvwTiles->Selected) return;

  TTiles& Tiles = TTiles::Instance();
  int No = (int)tvwTiles->Selected->Data; // oثeϮsսs

  grdTile->RowCount = Tiles.Tile[No].YNum; // ]w grid jp, ܾӹϮs
  grdTile->ColCount = Tiles.Tile[No].XNum;

  grdTile->Selection.Left = 0;
  grdTile->Selection.Top = 0;
  grdTile->Selection.Right = Tiles.Tile[No].XNum - 1;
  grdTile->Selection.Bottom = Tiles.Tile[No].YNum - 1;

  grdTile->Invalidate();
}
//---------------------------------------------------------------------------
// oܰϰ쪺eΰ
int __fastcall TMainForm::GetSelectionWidth()
{
  return grdTile->Selection.Right - grdTile->Selection.Left + 1;
}
//---------------------------------------------------------------------------
int __fastcall TMainForm::GetSelectionHeight()
{
  return grdTile->Selection.Bottom - grdTile->Selection.Top + 1;
}
//---------------------------------------------------------------------------

