//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#define tileunitCPP
#include "tileunit.h"
#include "util.h"

#include <typeinfo>

#pragma package(smart_init)

//---------------------------------------------------------------------------
TTile::TTile()
{
  init();
}

TTile::~TTile()
{
  delete FBits;
}

void TTile::init()
{
  // xϤ bitmap, jp TILE_WIDTH * TILE_HEIGHT
  FBits = new Graphics::TBitmap;
  FBits->Width = TILE_WIDTH;
  FBits->Height = TILE_HEIGHT;
  FBits->Transparent = true;
  FBits->TransparentColor = TRANSPARENT_COLOR;

  // w]iϤΦ@Ӹs
  FFirstTile = true;
  FDisposed = false;
  
  FXNum = 1;
  FYNum = 1;
}

void TTile::LoadFromStream(TStream* Stream)
{
  TReader* reader = new TReader(Stream, 2048);
  try {
    FDisposed = reader->ReadBoolean();
    reader->FlushBuffer();
    // O_wm ?? OܴNAŪ䥦ݩ
    if (FDisposed) return;

    FBits->LoadFromStream(Stream); // ϧ
    reader->FlushBuffer();

    // ݩ
    for (TTileAttrElement x = taCanPass; x <= taFlag; x = (TTileAttrElement)(x + 1)) {
      bool b = reader->ReadBoolean();
      if (b) FAttr = FAttr << x;
    }

    FFirstTile = reader->ReadBoolean();
    if (FFirstTile) { // YOsYY, hŪsժe
      FXNum = reader->ReadInteger();
      FYNum = reader->ReadInteger();
    }
  } __finally {
    delete reader;
  }
}

void TTile::SaveToStream(TStream* Stream)
{
  TWriter* writer = new TWriter(Stream, 2048);
  try {
    writer->WriteBoolean(FDisposed);
    writer->FlushBuffer();
    // O_wm ?? OܴNAgJ䥦ݩ
    if (FDisposed) return;

    FBits->SaveToStream(Stream); // ϧ

    writer->FlushBuffer();

    // ݩ
    for (TTileAttrElement x = taCanPass; x <= taFlag; x = (TTileAttrElement)(x + 1))
      writer->WriteBoolean(FAttr.Contains(x));

    writer->WriteBoolean(FFirstTile);
    if (FFirstTile) { // YOsYY, hgJsժe
      writer->WriteInteger(FXNum);
      writer->WriteInteger(FYNum);
    }
  } __finally {
    delete writer;
  }
}

TTile& TTile::operator=(const TTile& t)
{
  if (this != &t) { // קKۧڳ]
    FBits->Assign(t.FBits);
    FAttr = t.FAttr;
    FDisposed = t.FDisposed;

    FFirstTile = t.FFirstTile;
    FXNum = t.FXNum;
    FYNum = t.FYNum;
  }
  return *this;
}

TTile::TTile(const TTile& t)
{
  init();
  *this = t; // invoke assignment constructor
}

// TTiles class

TTiles* TTiles::FInstance = NULL;

TTiles& TTiles::Instance()
{
  if (FInstance == NULL) {
    FInstance = new TTiles;
  }
  return *FInstance;
}

TTiles::TTiles()
{
  FTiles = new TTileArray;
}

TTiles::~TTiles()
{
  FreeTiles();
  delete FTiles;
}

// [JsϤ, b}C̫
int TTiles::AddTile(TTile* NewTile)
{
  FTiles->push_back(NewTile);
  
  return FTiles->size() - 1;
}

// ҦϤ
void TTiles::FreeTiles()
{
  // n free Ҧ tile 
  for (TTileArray::iterator p = FTiles->begin(); p != FTiles->end(); p++)
    delete *p;

  FTiles->clear();
}

// qɮŪϮw
void TTiles::LoadFromFile(AnsiString FileName)
{
  TFileStream* fs = new TFileStream(FileName, fmOpenRead);
  TReader* reader = new TReader(fs, 2048);
  try {
    CheckSignature(reader, SIG_MYFILE); // ˬdY
    CheckSignature(reader, typeid(*this).name()); // ˬdY

    int num = reader->ReadInteger(); // ŪJ tiles ƥ
    if (!num) throw Exception("No tiles");

    reader->FlushBuffer();
    
    for (int i = 0; i < num; i++) {
      TTile* t = new TTile;
      t->LoadFromStream(fs); // ŪJҦ tiles

      FTiles->push_back(t);
    }
  } __finally {
    delete reader;
    delete fs;
  }
}

// gJϮw
void TTiles::SaveToFile(AnsiString FileName)
{
  TFileStream* fs = new TFileStream(FileName, fmCreate | fmOpenWrite);
  TWriter* writer = new TWriter(fs, 2048);
  try {
    writer->WriteString(SIG_MYFILE);
    writer->WriteString(typeid(*this).name());

    writer->WriteInteger((int)FTiles->size());
    writer->FlushBuffer();

    for (TTileArray::iterator p = FTiles->begin(); p != FTiles->end(); p++)
      (*p)->SaveToStream(fs); // gJҦ tiles
  } __finally {
    delete writer;
    delete fs;
  }
}

int TTiles::GetTileNum()
{
  return FTiles->size();
}

TTile& TTiles::GetTile(int No)
{
  return *(*FTiles)[No];
}


